@@ -52,6 +52,7 @@ class APIClient:
         new_keys: bool = False,
         delete_old_keys: bool = False,
         ssl_verify: bool = True,
+        cert_auth: bool = False
     ):
         """
         :param config: Config as a dictionary
@@ -60,9 +61,11 @@ class APIClient:
         """
         self.ssl_verify = ssl_verify
         self.config = config
+        self.cert_auth = cert_auth
+        self.cert = None
         if config_path:
             self.config = configparser.ConfigParser()
-            self.config.read_file(open(config_path, "r"))
+            self.config.read_file(open(config_path))
         self.requests_session = requests.Session()
 
         if not self.config:
@@ -71,6 +74,12 @@ class APIClient:
             raise ValueError("Missing value for SERVER in config.ini")
 
         self.server_url = self.config["PASSBOLT"]["SERVER"].rstrip("/")
+
+        if self.cert_auth:
+            if not (self.config["PASSBOLT"]["SERVER_CERT_AUTH_CRT"] and self.config["PASSBOLT"]["SERVER_CERT_AUTH_KEY"]):
+                raise ValueError("Missing certificate and key in config.ini")
+            self.cert = (self.config["PASSBOLT"]["SERVER_CERT_AUTH_CRT"], self.config["PASSBOLT"]["SERVER_CERT_AUTH_KEY"])
+        
         self.user_fingerprint = self.config["PASSBOLT"]["USER_FINGERPRINT"].upper().replace(" ", "")
         self.gpg = gnupg.GPG()
         if delete_old_keys:
@@ -107,14 +116,14 @@ class APIClient:
             raise ValueError("Missing value for USER_PUBLIC_KEY_FILE in config.ini")
         if not self.config["PASSBOLT"]["USER_PRIVATE_KEY_FILE"]:
             raise ValueError("Missing value for USER_PRIVATE_KEY_FILE in config.ini")
-        self.gpg.import_keys(open(self.config["PASSBOLT"]["USER_PUBLIC_KEY_FILE"], "r").read())
-        self.gpg.import_keys(open(self.config["PASSBOLT"]["USER_PRIVATE_KEY_FILE"], "r").read())
+        self.gpg.import_keys(open(self.config["PASSBOLT"]["USER_PUBLIC_KEY_FILE"]).read())
+        self.gpg.import_keys(open(self.config["PASSBOLT"]["USER_PRIVATE_KEY_FILE"]).read())
 
     def _login(self):
-        r = self.requests_session.post(self.server_url + LOGIN_URL, json={"gpg_auth": {"keyid": self.gpg_fingerprint}}, verify=self.ssl_verify)
+        r = self.requests_session.post(self.server_url + LOGIN_URL, json={"gpg_auth": {"keyid": self.gpg_fingerprint}}, verify=self.ssl_verify, cert=self.cert) # None is the default value in requests
         encrypted_token = r.headers["X-GPGAuth-User-Auth-Token"]
         encrypted_token = urllib.parse.unquote(encrypted_token)
-        encrypted_token = encrypted_token.replace("\+", " ")
+        encrypted_token = encrypted_token.replace(r"\+", " ")
         token = self.decrypt(encrypted_token)
         self.requests_session.post(
             self.server_url + LOGIN_URL,
@@ -122,16 +131,36 @@ class APIClient:
                 "gpg_auth": {"keyid": self.gpg_fingerprint, "user_token_result": token},
             },
         )
-        self._get_csrf_token()
+        try:
+            self._get_csrf_token()
+        except requests.exceptions.HTTPError as e:
+            if (
+                e.response.status_code != requests.status_codes.codes.forbidden
+                or e.response.json()["header"]["message"]
+                != "MFA authentication is required."
+            ):
+                logging.error(r.text)
+                raise e
+            if not self.config["PASSBOLT"]["OTP"]:
+                raise ValueError("Missing value for OTP in config.ini")
+            self.post("/mfa/verify/totp.json", {"totp": self.config["PASSBOLT"]["OTP"]})
 
     def _get_csrf_token(self):
-        self.get("/users/me.json", return_response_object=True)  # Fetches the X-CSRF-Token header for future requests
+        """Fetches the X-CSRF-Token header for future requests"""
+        r = self.requests_session.get(self.server_url + "/users/me.json")
+        r.raise_for_status()
 
     def encrypt(self, text, recipients=None):
         return str(self.gpg.encrypt(data=text, recipients=recipients or self.gpg_fingerprint, always_trust=True))
 
     def decrypt(self, text):
-        return str(self.gpg.decrypt(text, always_trust=True, passphrase=str(self.config["PASSBOLT"]["PASSPHRASE"])))
+        if "PASSPHRASE" in self.config["PASSBOLT"]:
+            passphrase = str(self.config["PASSBOLT"]["PASSPHRASE"])
+        else:
+            passphrase = None
+
+
+        return str(self.gpg.decrypt(text, always_trust=True, passphrase=passphrase))
 
     def get_headers(self):
         return {
@@ -210,7 +239,7 @@ class PassboltAPI(APIClient):
     def _get_secret(self, resource_id: PassboltResourceIdType) -> PassboltSecretTuple:
         response = self.get(f"/secrets/resource/{resource_id}.json")
         assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
-        return PassboltSecretTuple(**response["body"])
+        return constructor(PassboltSecretTuple)(response["body"])
 
     def _update_secret(self, resource_id: PassboltResourceIdType, new_secret):
         return self.put(f"/resources/{resource_id}.json", {"secrets": new_secret}, return_response_object=True)
@@ -250,8 +279,7 @@ class PassboltAPI(APIClient):
         response = self.get("/resources.json" + url_params)
         assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
         resources = response["body"]
-        for resource in resources:
-            yield resource
+        yield from resources
 
     def list_resources(self, folder_id: Optional[PassboltFolderIdType] = None):
         params = {
